/************************************************************
 * lzw2z.c : Reads the GRASS raster using G_lzw_read() and  *
 * then writes it using zlib compression                    *
 * -------------------------------------------------------- *
 * Author: Eric G. Miller Nov. 2000                         *
 * -------------------------------------------------------- *
 * This code is public domain                               *
 ***********************************************************/

#include "gis.h"
#include "local_proto.h"
#include <unistd.h>
#include <fcntl.h>
#include <zlib.h>
#include <assert.h>
#include <errno.h>

#define GZLIB_OK 0
#define GZLIB_EMEM (-1)
#define GZLIB_EREAD (-2)
#define GZLIB_ECMP (-3)
#define GZLIB_EWRITE (-4)
#define GZLIB_COMPRESSED_YES (unsigned char)'1'
#define GZLIB_COMPRESSED_NO  (unsigned char)'0'

#define XDR_FLOAT_NBYTES 4
#define XDR_DOUBLE_NBYTES 8

static int read_ptrs (int, long *, int);
static int write_ptrs (int, long *, int);
static int test_inflate (int, RASTER_MAP_TYPE, int, int);

static unsigned char *
myalloc (ptr, sz)
    unsigned char *ptr;
    int  sz;
{
    unsigned char *new;

    if (ptr == NULL)
        new = (unsigned char *) G_malloc (sz);
    else
        new = (unsigned char *) G_realloc (ptr, sz);
    
    return new;
}

static void
cleanup (src, dst, z)
    unsigned char *src, *dst;
    z_stream *z;
{
    if (src != NULL)
        G_free(src);
    if (dst != NULL)
        G_free(dst);
    if (z != NULL)
        deflateEnd(z);
}


int
lzw2z (fd_in, head, map_type, fd_out)
    int fd_in, fd_out;
    struct Cell_head head;
    RASTER_MAP_TYPE map_type;
{
    int err, row, nbytes, nread;
    unsigned char *src, *dst, compressed;
    z_stream c_stream;
    int src_sz, dst_sz;
    long *lz_offset, *z_offset;
    
    src = dst = NULL;
    switch (map_type)
    {
        case FCELL_TYPE: 
            nbytes = XDR_FLOAT_NBYTES; 
            break;
        case DCELL_TYPE: 
            nbytes = XDR_DOUBLE_NBYTES; 
            break;
        default:
            G_fatal_error ("lzw2z [%d]: Unknown file format %d\n", 
                    __LINE__, head.format);
            break;
    }

    /* We should never need more than this size of bytes */
    src_sz = nbytes * head.cols;
    /* dst needs to be bigger for single pass deflate */
    dst_sz = (int) src_sz * 1.1 + 100;
    
    src = myalloc (src, src_sz);
    if (src == NULL)
        return GZLIB_EMEM;
    dst = myalloc (dst, dst_sz);
    if (dst == NULL)
    {
        G_free(src);
        return GZLIB_EMEM;
    }

    /* Setup zstream and initialize */
    c_stream.zalloc = (alloc_func)0;
    c_stream.zfree  = (free_func)0;
    c_stream.opaque = (voidpf)0;

    /* Get lzw row offsets and write preliminary z row offsets */
    lz_offset = (long *) G_calloc (head.rows + 1, sizeof(long));
    if (lz_offset == NULL)
    {
        G_free(src); G_free(dst);
        return GZLIB_EMEM;
    }
    z_offset = (long *) G_calloc (head.rows + 1, sizeof(long));
    if (z_offset == NULL)
    {
        G_free(src); G_free(dst); G_free(lz_offset);
        return GZLIB_EMEM;
    }
    if (read_ptrs (fd_in, lz_offset, head.rows) == -1)
        return GZLIB_EREAD;
    for (row = 0; row < head.rows + 1; row++)
        z_offset[row] = lz_offset[row];
    if ((z_offset[0] = (long) write_ptrs 
                (fd_out, z_offset, head.rows)) < head.rows)
        return GZLIB_EWRITE;
    
    nread = 0;
    for (row = 0; row < head.rows; row++)
    {
        assert (src != NULL);
        /* Do input row uncompress, use offsets to set file position */
        lseek (fd_in, lz_offset[row], SEEK_SET);
        nbytes = G_lzw_read (fd_in, src, src_sz);
        if (nbytes == 0) /* No more bytes?? */
        {
            cleanup (src, dst, &c_stream);
            return GZLIB_EREAD;
        }
        if (nbytes == -1) /* Not enough space ??? */
        {
            cleanup (src, dst, &c_stream);
            return GZLIB_EREAD;
        }
        if (nbytes == -2)
        {
            cleanup (src, dst, &c_stream);
            return GZLIB_EREAD;
        }
        if (nbytes < -2) /* Shouldn't happen */
        {
            cleanup (src, dst, &c_stream);
            return GZLIB_EREAD;
        }
        nread += G_lzw_nof_read_bytes ();
        
        /* Do row deflate */
        c_stream.next_in = src;
        c_stream.avail_in = nbytes;
        c_stream.next_out = dst;
        c_stream.avail_out = dst_sz;
        assert (src != NULL && dst != NULL);
        err = deflateInit (&c_stream, Z_DEFAULT_COMPRESSION);
        if (err != Z_OK)
        {
            G_free (src);
            G_free (dst);
            switch (err)
            {
                case Z_STREAM_ERROR:
                case Z_VERSION_ERROR:
                    return GZLIB_ECMP;
                    break;
                case Z_MEM_ERROR:
                    return GZLIB_EMEM;
                    break;
            }
        }
        err = deflate (&c_stream, Z_FINISH);

        if (err != Z_STREAM_END)
        {
            cleanup (src, dst, &c_stream);
            return GZLIB_ECMP;
        }

        /* Check that the compressed size is smaller than input */
        if (nbytes < c_stream.total_out)
        {
            /* It's not, so write out uncompressed row */
            lseek (fd_out, z_offset[row], SEEK_SET);
            compressed = GZLIB_COMPRESSED_NO;
            /* Write compression flag */
            if (write (fd_out, &compressed, 1) != 1)
            {
                cleanup (src, dst, &c_stream);
                return GZLIB_EWRITE;
            }
            /* Write data */
            nread = nbytes;
            nbytes = 0;
            do {
                err = write (fd_out, dst + nbytes, nread - nbytes);
                if (err >= 0)
                    nbytes += err;
            } while (nbytes < nread && err > 0);

            /* If there was an error writing */
            if (err < 0)
            {
                cleanup (src, dst, &c_stream);
                return GZLIB_EWRITE;
            }
            
            /* update z_offset */
            z_offset[row+1] = lseek (fd_out, 0L, SEEK_CUR);
            deflateEnd (&c_stream);
        }
        else /* Compression made it smaller */
        {
            /* Write the compressed row */
            err = 1;
            nbytes = 0;
            lseek (fd_out, z_offset[row], SEEK_SET);
            /* Write compression flag */
            compressed = GZLIB_COMPRESSED_YES;
            if (write (fd_out, &compressed, 1) != 1)
            {
                cleanup (src, dst, &c_stream);
                return GZLIB_EWRITE;
            }
            while (err > 0)
            {
                err = write (fd_out, dst + nbytes, c_stream.total_out - nbytes);
                if (err >= 0)
                    nbytes += err;
            }
            if (err < 0 && nbytes != c_stream.total_out)
            {
                cleanup (src, dst, &c_stream);
                return GZLIB_EWRITE;
            }
            /* update z_offset */
            z_offset[row + 1] = lseek (fd_out, 0L, SEEK_CUR);
            deflateEnd (&c_stream);
        } /* if (nbytes <  c_stream.total_out) */
    } /* bottom of for(;;) */

    /* Write z_offsets again */
    write_ptrs (fd_out, z_offset, head.rows);
    /* Clean Up */
    cleanup (src, dst, &c_stream);
    G_free (lz_offset);
    G_free (z_offset);

    /* test inflate */
    test_inflate (fd_out, map_type, head.rows, head.cols);
    
    return GZLIB_OK;
}



static int read_ptrs (fd, ptrs, rows)
    int fd, rows;
    long *ptrs;
{
    int n ;
    int row;
    unsigned char *buf, *b;
    unsigned char nbytes[1];
    long v;
    
/*
 * 3.0 row address array is in a machine independent format
 * (warning - the format will work even if the sizeof(long) is
 *  not the same from machine to machine, as long as the
 *  actual values do not exceed the capability of the long)
 */
    if (read (fd, nbytes, 1) != 1)
    {
	goto badread;
    }
    if (nbytes[0] == 0)
    {
	goto badread;
    }
    n = (rows + 1) * nbytes[0] ;
    b = buf = (unsigned char *) G_malloc (n);
    if (read (fd, b, n) != n)
    {
	goto badread;
    }
    for (row = 0; row <= rows; row++)
    {
	v = 0;
	for (n = 0; n < (int) nbytes[0]; n++)
	    v = (v << 8) + *b++;   /* v = (v * 256) + *b++; */
	ptrs[row] = v;
    }
    G_free(buf);
    return 1;

badread:
    G_warning ( "Fail of initial read of compressed file") ;
    return -1;
}

static int write_ptrs (fd, ptrs, rows)
    int fd, rows;
    long *ptrs;
{
    int i,len, nbytes;
    int row;
    long x, y;
    unsigned char *buf, *b;

    lseek (fd, 0L, SEEK_SET);
    len = (rows+1) * sizeof(long) + 1;
    b = buf = (unsigned char *) G_malloc (len);
    *b++ = sizeof(long);

    for (row = 0; row <= rows; row++)
    {
	x = ptrs[row];
	i = sizeof(long);
	while (i-- > 0)
	{
	    y = x >> 8;           /* y = x / 256; */
	    b[i] = x - (y << 8);  /* b[i] = x - (y * 256); */
	    x = y;
	}
	b += sizeof(long);
    }

    nbytes = 0;
    while (nbytes < len && (i = write (fd, buf+nbytes, len-nbytes)) > 0)
        if (i >= 0)
            nbytes += i;
    G_free(buf);

    if (i < 0)
        return i;
    else
        return nbytes;
}

static int 
test_inflate (fd, map_type, rows, cols)
    int fd, rows, cols;
    RASTER_MAP_TYPE map_type;
{
    int  row, nbytes, nalloc, err;
    unsigned char *src, *dst, compressed;
    long *index, src_sz, dst_sz;
    z_stream c_stream;

    /* Compute uncompressed bytes per row */
    switch (map_type)
    {
        case FCELL_TYPE:
            nbytes = cols * XDR_FLOAT_NBYTES; break;
        case DCELL_TYPE:
            nbytes = cols * XDR_DOUBLE_NBYTES; break;
        default:
            G_fatal_error ("Unknown raster type %d", map_type);
            break;
    }

    nalloc = nbytes;
    /* Allocate memory */
    index = (long *) G_calloc (rows + 1, sizeof(long));
    if (index == NULL)
        G_fatal_error ("test_inflate [%d]: G_calloc", __LINE__);
    src = myalloc (NULL, nalloc);
    if (src == NULL)
        G_fatal_error ("test_inflate [%d]: G_calloc", __LINE__);
    dst = myalloc (NULL, nalloc);
    if (dst == NULL)
        G_fatal_error ("test_inflate [%d]: G_calloc", __LINE__);
    src_sz = dst_sz = nalloc;
    
    /* Get row index */
    lseek (fd, 0L, SEEK_SET);
    if (read_ptrs (fd, index, rows) == -1)
        G_fatal_error ("test_inflate [%d]: read_ptrs() failed", __LINE__);

    /* Initialize z_stream structure */
    c_stream.zalloc = (alloc_func)0;
    c_stream.zfree  = (free_func)0;
    c_stream.opaque = (voidpf)0;

    /* Begin reading */
    for (row = 0; row < rows; row++)
    {
        lseek (fd, index[row], SEEK_SET);
        src_sz = index[row + 1] - index[row];
        assert (src_sz <= nalloc + 1);
        /* read compressed flag */
        if (read (fd, &compressed, 1) != 1)
            G_fatal_error ("test_inflate [%d]: read()", __LINE__);
        if (compressed == GZLIB_COMPRESSED_NO)
        {
            if (read (fd, src, src_sz - 1) != src_sz - 1)
                G_fatal_error ("test_inflate [%d]: read()", __LINE__);
            /* Nothing else to do, hit the continue button */
            continue;
        }
        else if (compressed != GZLIB_COMPRESSED_YES)
        {
            /* Where the heck are we in the file?? */
            G_fatal_error ("test_inflate [%d]: read() -- Index Error",
                    __LINE__);
        }
        /* Okay, it's a compressed row */
        if (read (fd, src, src_sz - 1) != src_sz - 1)
            G_fatal_error ("test_inflate [%d]: read()", __LINE__);
            
        c_stream.next_in = src;
        c_stream.avail_in = nalloc;
        c_stream.next_out = dst;
        c_stream.avail_out = dst_sz;
        
        err = inflateInit (&c_stream);
        switch (err)
        {
            case Z_OK:
                break;
            case Z_MEM_ERROR:
                G_fatal_error ("test_inflate [%d]: inflateInit() -- Memory",
                        __LINE__);
                break;
            case Z_VERSION_ERROR:
                G_fatal_error ("test_inflate [%d]: inflateInit() -- Version",
                        __LINE__);
            default:
                break;
        }

        err = inflate (&c_stream, Z_FINISH);
        switch (err)
        {
            case Z_STREAM_END:
                assert (0 == c_stream.avail_out);
                break;
            case Z_OK:
                G_fatal_error ("test_inflate [%d]: row %d inflate() incomplete",
                        __LINE__, row);
                break;
            case Z_STREAM_ERROR:
                G_fatal_error ("test_inflate [%d]: row %d "\
                        "inflate() Z_STREAM_ERROR", __LINE__, row);
                break;
            case Z_DATA_ERROR:
                G_fatal_error ("test_inflate [%d]: row %d "\
                        "inflate() Z_DATA_ERROR", __LINE__, row);
                break;
            case Z_MEM_ERROR:
                G_fatal_error ("test_inflate [%d]: row %d "\
                        "inflate() Z_MEM_ERROR", __LINE__, row);
                break;
            case Z_BUF_ERROR:
                G_fatal_error ("test_inflate [%d]: row %d "\
                        "inflate() Z_BUF_ERROR", __LINE__, row);
                break;
            default:
                G_fatal_error ("test_inflate [%d]: row %d "\
                        "inflate() Unknown error %d", __LINE__, row, err);
                break;
        } /* switch (err) */
        inflateEnd (&c_stream);
    } /* for (row = ... ) */

    G_free (src);
    G_free (dst);
    G_free (index);

    return 0;
} /* test_inflate() */
    
/* vim: set softtabstop=4 shiftwidth=4 expandtab: */
