/***********************************************************
 * r.lzw2z -- module to convert GRASS rasters from LZW     *
 * compression to deflate compression                      *
 * ------------------------------------------------------- *
 * Author: Eric G. Miller Nov. 2000 (main.c lzw2z.c)       *
 *         Unknown Author and Date  (lzw.c)                *
 *                                                         *
 * This software is released into the public domain.  The  *
 * lzw.c code was originally part of the GRASS codebase    *
 * released into the public domain by USA Army CERL.       *
 **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <dirent.h>
#include <string.h>
#include "gis.h"
#include "local_proto.h"


struct my_rasters {
    char *name;
    char *mapset;
    struct my_rasters *next;
};

static void
warn_user (void)
{
    int i, yes_no = NO;
    const char msg[] = "\n"\
	"\tWARNING: This is experimental code. While it makes every\n"\
	"\tattempt to insure your files are properly translated, no\n"\
	"\tguarantee is made.\n\n"\
	"\tPLEASE backup your files before running this operation.\n\n";
    const char prompt[] = "continue? (y/n) > ";
    
    fprintf (stdout, "%s", msg);

    for (i = 0; i < 3; i++)
    {
	int c;
	fprintf (stdout, "%s", prompt);
        fflush(stdout);
	c = getchar();
	if (c == 'y' || c == 'Y') 
	{
	    yes_no = YES;
	    break; /* out of for(;;) */
	}
	if (c == 'n' || c == 'N')
	{
	    yes_no = NO;
	    break; /* out of for(;;) */
	}
	while (c != EOF && c != '\n')
            c = getchar(); /* Eat rest of input */
    }

    if (yes_no == NO) {
	fprintf (stdout, "Aborted!\n");
	exit (EXIT_SUCCESS);
    }

    fprintf (stdout, "Doing translations ...\n");
}


int
main (int argc, char **argv)
{
    int fd_in, fd_out, all, confirm, err, len;
    char *name, *mapset, *location, *thedir, *dummy1, *dummy2;
    struct Flag	    *do_all, *no_warn;
    struct Option   *cell_opt;
    struct my_rasters theList, *this;
    struct Cell_head head;
    struct GModule *module;
    
    G_gisinit(argv[0]);

    module              = G_define_module();
    module->description = "Convert compressed floating point rasters from "\
        "LZW compression to Deflate.";
    
    cell_opt		= G_define_option ();
    cell_opt->key	= "rast";
    cell_opt->required	= NO;
    cell_opt->multiple	= YES;
    cell_opt->gisprompt = "old,cell,raster";
    cell_opt->description = "One or more rasters to translate";

    do_all		= G_define_flag ();
    do_all->key		= 'a';
    do_all->description = "Process all rasters in current location";

    no_warn		= G_define_flag ();
    no_warn->key        = 'f';
    no_warn->description = "Do not warn about this DANGEROUS OPERATION.\n"\
	"And, do not prompt for confirmation";

    if (G_parser(argc, argv) != 0)
	exit (EXIT_FAILURE);

    all = NO;
    confirm = YES;
    
    if (do_all->answer)
    {
	all = YES; 
    }

    if (no_warn->answer)
    {
	confirm = NO;
    };

    if (all == NO && (cell_opt->answer == NULL 
		&& cell_opt->answers == NULL))
    {
	/* Nothing to do */
	G_fatal_error ("You must specify at least one raster or use "
		"the '-a' flag");
    }

     /* Get the current GRASS mapset path */
    location = G_location_path();
    mapset   = G_mapset();
    if (location == NULL || mapset == NULL)
        G_fatal_error ("Unable to get location and mapset");

    len = strlen(location) + strlen(mapset) + 8;
    thedir = G_calloc (len, sizeof(char));
    if (thedir == NULL)
        G_fatal_error ("Memory Allocation Failed");
    snprintf (thedir, len, "%s/%s/fcell", location, mapset);
    
    if (all == YES)
    {
        int first = 1;
        DIR *dir;
        struct dirent *dp;
        
       
        /* Open and Read directory contents */
        dir = opendir(thedir);
        if (dir == NULL)
            G_fatal_error ("Unable to open directory %s\n"\
                    "No Floating Point rasters in mapset?", thedir);
        while (NULL != (dp = readdir (dir)))
        {
            /* Weed out '.' and '..' */
            if (strncmp (dp->d_name, ".", 1) == 0)
                continue;

            if (first)
            {
                len = strlen (dp->d_name) + 1;
                theList.name = G_calloc (len, sizeof(char));
                if (theList.name == NULL)
                    G_fatal_error ("Memory Allocation Failed");
                strncpy (theList.name, dp->d_name, len - 1);
                theList.mapset = mapset;
                theList.next = NULL;
                this = &theList;
                first = 0;
            }
            else
            {
		this->next = (struct my_rasters *)
			G_malloc (sizeof(struct my_rasters));
		if (this->next == NULL)
		    G_fatal_error ("G_malloc failed for list of rasters");
		this = this->next;
                len = strlen (dp->d_name) + 1;
                this->name = G_calloc (len, sizeof(char));
                if (this->name == NULL)
                    G_fatal_error ("Memory Allocation Failed");
                strncpy (this->name, dp->d_name, len - 1);
		this->mapset = mapset;
		this->next = NULL;
	    } /* if (first) */
        } /* while (NULL != (dp = readdir (dir))) */
        closedir (dir);
    } /* if (all == YES) */
    else if (cell_opt->answers != NULL)
    {
	assert(*cell_opt->answers != NULL);
	for (name = *cell_opt->answers; name != NULL ; )
	{
	    int first = 1;

	    mapset = G_find_file ("cell", name, "");
	    if (mapset == NULL)
		G_fatal_error ("Raster %s not found", name);
	    if (first)
	    {
		theList.name = name;
		theList.mapset = mapset;
		theList.next = NULL;
		this = &theList;
	    }
	    else
	    {
		this->next = (struct my_rasters *)
			G_malloc (sizeof(struct my_rasters));
		if (this->next == NULL)
		    G_fatal_error ("G_malloc failed for list of rasters");
		this = this->next;
		this->name = name;
		this->mapset = mapset;
		this->next = NULL;
	    }
            cell_opt->answers++;
            name = *cell_opt->answers;
	}
    }
    else if (cell_opt->answer != NULL)
    {
	name = cell_opt->answer;
	mapset = G_find_file ("cell", name, "");
	if (mapset == NULL)
	    G_fatal_error ("Raster %s not found", name);
	theList.name = name;
	theList.mapset = mapset;
	theList.next = NULL;
    }
    else
    {
	G_fatal_error ("Programmer error in making list of rasters");
    }

    /* Warn the user about the big danger unless overridden */
    if (confirm)
	warn_user();
    
    /* These are used by G_is_reclass() */
    dummy1 = G_malloc (80);
    dummy2 = G_malloc (80);
    if (dummy1 == NULL || dummy2 == NULL)
        G_fatal_error ("main.c [%d]: Memory Allocation Failed", __LINE__);
    
    /* Now we do it */
    this = &theList;
    len = strlen(thedir);
    while (this != NULL)
    {
        int pathlen;
	char *tmpname, *path = NULL;
        RASTER_MAP_TYPE maptype;

        /* If it's a CELL_TYPE or FP and not compressed, skip it */
        if ((maptype = G_raster_map_type (this->name, this->mapset))
                == CELL_TYPE)
        { 
            fprintf (stderr, "Skipping CELL raster %s@%s\n",
                    this->name, this->mapset);
            this = this->next;
            continue;
        }
        if (!(maptype == DCELL_TYPE || maptype == FCELL_TYPE))
        {
            fprintf (stderr, "Skipping unknown raster type %s@%s\n",
                    this->name, this->mapset);
            fprintf (stderr, "Got maptype %d, only know:\n"\
                    "CELL_TYPE: %d\nFCELL_TYPE: %d\nDCELL_TYPE: %d\n",
                    maptype, CELL_TYPE, FCELL_TYPE, DCELL_TYPE);
            this = this->next;
            continue;
        }
        if (G_get_cellhd (this->name, this->mapset, &head) != 0)
            G_fatal_error ("Couldn't open cell head for %s@%s",
                    this->name, this->mapset);
        if (head.compressed == 0)
        {
            fprintf (stderr, "Skipping uncompressed raster %s@%s\n",
                    this->name, this->mapset);
            this = this->next;
            continue;
        }

        /* If G_mapset() != this->mapset, we can't do update */
        if (strncmp (G_mapset(), this->mapset, strlen (this->mapset))
                != 0)
        {
            fprintf (stderr, "Skipping %s@%s, not in current mapset %s\n",
                    this->name, this->mapset, G_mapset());
            this = this->next;
            continue;
        }
        
        /* We don't want to operate on reclass maps either */
        if (G_is_reclass (this->name, this->mapset, dummy1, dummy2) != 0)
        {
            fprintf (stderr, "Skipping reclass raster %s@%s of %s@%s\n",
                    this->name, this->mapset, dummy1, dummy2);
            this = this->next;
            continue;
        }

        /* Make sure lzw_compression_bits != -1 (already processed) */
        if (!islzw (this->name, this->mapset))
        {
            fprintf (stderr, "Skipping raster %s@%s -- Not LZW compressed\n",
                    this->name, this->mapset);
            this = this->next;
            continue;
        }
        
        /* Okay, it's FP *and* compressed */
        /* Realloc *path if necessary */
        pathlen = len + strlen(this->name) + 2;
        if (NULL == (path = G_realloc(path, pathlen)))
            G_fatal_error ("%s [%d]: G_realloc failed!", __FILE__, __LINE__);
        snprintf (path, pathlen, "%s/%s", thedir, this->name);
	if ((fd_in = open (path, O_RDONLY)) < 0)
	    G_fatal_error ("%s [%d]: Couldn't open %s@%s", __FILE__, __LINE__,
                    this->name, this->mapset);
	tmpname = G_tempfile();
	if ((fd_out = open (tmpname, O_RDWR | O_CREAT | O_TRUNC, 0644)) < 0)
	    G_fatal_error ("Couldn't open temporary file %s", tmpname);
        fprintf (stderr, "\nProcessing %s@%s ... ", this->name, this->mapset);
        if ((err = lzw2z (fd_in, head, maptype, fd_out)) != 0)
        {
	    fprintf (stderr, "Error doing translations for %s@%s. ",
		    this->name, this->mapset);
            if (err == -1)
                fprintf (stderr, "MEMORY\n");
            else if (err == -2)
                fprintf (stderr, "READ\n");
            else if (err == -3)
                fprintf (stderr, "COMPRESSION\n");
            else if (err == -4)
                fprintf (stderr, "WRITE\n");
            else
                fprintf (stderr, "UNKNOWN\n");
            exit (EXIT_FAILURE);
        }
        else
        {
            fprintf (stderr, "Copying ... ");
        }
        /* Close input file, truncate it, then rewrite it */
        close (fd_in);
        if (0 > (fd_in = open (path, O_WRONLY, 0644)))
            G_fatal_error ("Failed to open %s@%s for rewriting", this->name,
                    this->mapset);
        if (ftruncate(fd_in, 0L) != 0)
        {
            close(fd_in);
            G_fatal_error ("Truncate of %s@%s failed, file may be damaged",
                    this->name, this->mapset);
        }
        
        /* Rewind fd_out */
        lseek (fd_out, 0L, SEEK_SET);

        do {
            unsigned char b;
            err = read (fd_out, &b, 1);
            if (err != 1)
                break;
            err = write (fd_in, &b, 1);
        } while (err > 0);
          
        if (err == -1)
            G_fatal_error ("There was an error transfering the new raster to"\
                    " %s@%s", this->name, this->mapset);
        
        /* Set lzw_compression_bits = -1 (an invalid value) */
        if (setlzwbits (this->name, this->mapset, maptype) != 0)
            G_warning ("cell_misc/%s/f_format file may be corrupted!", 
                    this->name);

        /* Done with the translation */
	close (fd_in);
	close (fd_out);
        fprintf (stderr, "Done!\n");

	unlink (tmpname);
	G_free (tmpname);
	this = this->next;
    } /* while (this != NULL) */
    
    return 0;

}
/* vim: set softtabstop=4 shiftwidth=4 expandtab: */
